--------------------------------------------------------------------------------
-- Company: 
-- Engineer:
--
-- Create Date:    13:18:09 02/03/06
-- Design Name:    
-- Module Name:    FLASH - Behavioral
-- Project Name:   
-- Target Device:  
-- Tool versions:  
-- Description:
--
-- At power on, this chip automatically starts FPGA configuration from a FLASH memory.
-- After that local bus access is enabled.
--
-- It is possible to program the FPGA directly using slave parellel mode.
-- To do that, first write x"03" to address "10xxxxx", then write x"01" to the same address.
-- read data from address "1xxxxxx" until bit 2 is '1'
-- writing configuration byte data to address "00xxxxx" until all data done
-- 
-- The FLASH memory is programmed in blocks of 256 bytes.
-- To program a FLASH block, first write x"01" to address '0010000" to reset the buffer address counter.
-- Next write 256 bytes of data to address "0010001".
-- FLASH address is set by writing to addresses "1100000"(LSB byte), "1100001"(middle byte) and
-- "1100010"(two MSB bits)
-- To start FLASH programming, First set FLASH address to x"5555", then write x"AA" to address "01xxxxx"
-- Next set FLASH address to x"2AAA", then write x"55" to address "01xxxxx"    
-- Next set FLASH address to x"5555", then write x"A0" to address "01xxxxx"    
-- Finally set FLASH address to the starting address of the block,
-- When FPGA sees that FLASH address has been wriiten, it brings ctrl_in(1) low and starts sending
-- data to the FLASH. After 256 bytes being sent, FPGA brings ctrl_in(1) high again to enable the local bus   
-- Dependencies:
-- 
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
-- 
--------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
library UNISIM;
use UNISIM.VComponents.all;

entity HTR_SLB is
    Port ( 
           loc_add : in std_logic_vector(6 downto 0);
           loc_add_out : out std_logic_vector(6 downto 0);
           loc_bus : inout std_logic_vector(7 downto 0);
           loc_rw : in std_logic;
           loc_cs : in std_logic;
           maddr : out std_logic_vector(18 downto 0);
           mdata : inout std_logic_vector(7 downto 0);
           moe_b : out std_logic;
           mce_b : out std_logic;
           mwe_b : out std_logic;
           cclk_fb : in std_logic;
           cclk_fbp : out std_logic;
           cclk : in std_logic;
           cclkp : out std_logic;
			  prog_b : out std_logic;
			  cs_b : out std_logic;
			  spare5 : out std_logic;
			  rdwr_b : out std_logic;
           init_b : in std_logic;
           done : in std_logic;
           ctrl_in : in std_logic_vector(2 downto 0)
			  );
end HTR_SLB;

architecture Behavioral of HTR_SLB is
constant version : std_logic_vector(7 downto 0) := x"01";
constant prog_addr : std_logic_vector(6 downto 0) := "1000011";
constant sel_flash : std_logic_vector(6 downto 0) := "0100000";
constant sel_FPGA : std_logic_vector(6 downto 5) := "00";
constant sel_maddr : std_logic_vector(6 downto 2) := "11000";
signal maddr_cntr : std_logic_vector(17 downto 0) := (others => '0');
signal sel_byte : std_logic_vector(1 downto 0);
signal d_powerup : std_logic;
signal powerup : std_logic;
signal prog : std_logic;
signal cfg_done : std_logic;
signal loc_mode : std_logic;
signal w_data : std_logic_vector(7 downto 0);
signal r_data : std_logic_vector(7 downto 0);
signal en_loc_bus : std_logic := '1';
signal wr_maddr : std_logic;
signal en_conf : std_logic;
signal d_en_conf : std_logic;
signal d_en_loc_bus : std_logic;
signal cpld_data : std_logic_vector(7 downto 0);
signal crc : std_logic_vector(23 downto 0);
signal ring_os : std_logic_vector(3 downto 0);
signal div : std_logic_vector(4 downto 0);
signal div_n : std_logic_vector(4 downto 0);
signal rst_div : std_logic;
attribute noreduce : string;
attribute noreduce of ring_os : signal is "true";
attribute init : string;
attribute init of i_en_conf : label is "0";
attribute init of i_en_loc_bus : label is "0";
attribute init of i_powerup : label is "1";
attribute pwr_mode : string;
attribute pwr_mode of cclk : signal is "std";
begin
prog_b <= '0' when prog = '1' or powerup = '1' else 'Z';
loc_add_out <= loc_add;
i_g: for i in 1 to 3 generate
  i_buf: buf port map(i => ring_os(i-1), o => ring_os(i));
end generate;
ring_os(0) <= not cclk_fb or loc_mode or cfg_done or (not init_b and not powerup);
i_cclk_fb: obuf port map(i => ring_os(3), o => cclk_fbp);
rst_div <= not init_b and not powerup;
i_div0 : FDCPE
   generic map (
     INIT => '0') -- Initial value of register ('0' or '1')  
   port map (
      Q => div(0),      -- Data output
      C => cclk_fb,      -- Clock input
      CE => '1',    -- Clock enable input
      CLR => rst_div,  -- Asynchronous clear input
      D => div_n(0),      -- Data input
      PRE => '0'   -- Asynchronous set input
   );
g_div: for i in 1 to 4 generate	
  i_div : FDCPE
   generic map (
     INIT => '0') -- Initial value of register ('0' or '1')  
   port map (
      Q => div(i),      -- Data output
      C => div_n(i-1),      -- Clock input
      CE => '1',    -- Clock enable input
      CLR => rst_div,  -- Asynchronous clear input
      D => div_n(i),      -- Data input
      PRE => '0'   -- Asynchronous set input
   );
end generate;
div_n <= not div;	
process(cfg_done,ctrl_in,div,loc_mode,loc_cs,loc_rw,loc_add)
variable sel : std_logic_vector(1 downto 0);
begin
  sel := cfg_done & loc_mode;
  case sel is
    when "00" => cclkp <= not div(4);
    when "01" => cclkp <= loc_cs or loc_rw or loc_add(6) or loc_add(5);
    when others => cclkp <= ctrl_in(0) and (not ctrl_in(1) or loc_cs);
  end case;
end process;
r_data <= mdata when loc_add(6) = '0' else cpld_data;
process(loc_add,crc,ctrl_in,done,init_b,prog,loc_mode)
variable sel : std_logic_vector(2 downto 0);
begin
  sel := loc_add(5) & loc_add(1 downto 0);
  case sel is
    when "000" =>
		cpld_data <= crc(7 downto 0);
    when "001" =>
		cpld_data <= crc(15 downto 8);
    when "010" =>
		cpld_data <= crc(23 downto 16);
    when "011" =>
		cpld_data <= '0' & ctrl_in & done & init_b & prog & loc_mode;
    when "100" =>
		cpld_data <= maddr_cntr(7 downto 0);
    when "101" =>
		cpld_data <= maddr_cntr(15 downto 8);
    when "110" =>
		cpld_data <= "000000" & maddr_cntr(17 downto 16);
    when others =>
		cpld_data <= version;
  end case;
end process;  
loc_bus <= r_data when loc_cs = '0' and loc_rw = '1' and en_loc_bus = '1' else (others => 'Z');
mdata <= w_data when loc_rw = '0' and ((done = '1' and ctrl_in(1) = '1') or (done = '0' and loc_mode = '1')) else (others => 'Z');
process(loc_cs)
begin
  if(loc_cs'event and loc_cs = '0')then
    if(loc_rw = '0' and loc_add = prog_addr and en_loc_bus = '1')then
-- have to reset prog to start configuration
      prog <= loc_bus(1);
      loc_mode <= loc_bus(0);
	 end if;
	 w_data <= loc_bus;
	 sel_byte <= loc_add(1 downto 0);
  end if;
end process;
process(loc_cs,ctrl_in)
begin
  if(ctrl_in(1) = '0')then
    wr_maddr <= '0';
  elsif(loc_cs'event and loc_cs = '0')then
    if(loc_add(6 downto 2) = sel_maddr and loc_rw = '0' and cfg_done = '1')then
      wr_maddr <= '1';
	 else
      wr_maddr <= '0';
	 end if;	
  end if;
end process;
--wr_maddr <= '1' when loc_add(6 downto 2) = sel_maddr and loc_rw = '0' else '0';
--wr_maddr <= '1' when loc_add(6 downto 2) = sel_maddr and loc_rw = '0' and ctrl_in(1) = '1' and done = '1' else '0';
-- ctrl_in(1) is FLASH write enable
--moe_b <= done and not ctrl_in(2);
moe_b <= cfg_done and (not ctrl_in(1) or not loc_rw);
-- ctrl_in(0) is the low active FLASH chip enable signal
-- mce_b <= (done or loc_mode) and ctrl_in(0);
mce_b <= '0' when (cfg_done = '0' and loc_mode = '0') or ctrl_in(0) = '0' or
       (ctrl_in(1) = '1' and loc_add = sel_flash and loc_cs = '0') else '1';
--mwe_b <=	not done or ctrl_in(2);
mwe_b <=	not cfg_done or (ctrl_in(1) and loc_rw);
-- rdwr_b stays low when FPGA is not configured
-- when done is high, it enables FPGA's io_bus output when high 
rdwr_b <= cfg_done and loc_rw;
cs_b <= (not cfg_done and not (init_b and en_conf)) or (cfg_done and loc_cs);
spare5 <= cfg_done and loc_cs;
maddr <= '0' & maddr_cntr;
process(cclk, init_b)
begin
  if(init_b = '0' and powerup = '0')then
	 maddr_cntr <= (others => '0');
	 cfg_done <= '0';
	 crc <= (others => '1');
  elsif(cclk'event and cclk = '1')then
    if(wr_maddr = '1' and cfg_done = '1')then
      case sel_byte is
	     when "00" => maddr_cntr(7 downto 0) <= w_data;
	     when "01" => maddr_cntr(15 downto 8) <= w_data;
	     when "10" => maddr_cntr(17 downto 16) <= w_data(1 downto 0);
	     when others => maddr_cntr(17 downto 0) <= "000000000000000000";
	   end case;
    elsif(ctrl_in(1) = '0' or cfg_done = '0')then
      maddr_cntr(7 downto 0) <= maddr_cntr(7 downto 0) + 1;
		if(maddr_cntr(7 downto 0) = x"ff")then
		  maddr_cntr(17 downto 8) <= maddr_cntr(17 downto 8) + 1;
		end if;
	 end if;
	 if(maddr_cntr = "110011110110100111" and done = '1')then
	   cfg_done <= '1';
	 end if;
--	 if(init_b = '1' and done = '0' and maddr_cntr < "110011110110101000" )then
	 if(cfg_done = '0' and en_conf = '1')then
	   crc(0) <= mdata(0) xor crc(21) xor crc(20) xor crc(19) xor crc(18) xor crc(17) xor crc(16);
	   crc(1) <= mdata(1) xor crc(22) xor crc(16);
	   crc(2) <= mdata(2) xor crc(23) xor crc(17);
	   crc(3) <= mdata(3) xor crc(21) xor crc(20) xor crc(19) xor crc(17) xor crc(16);
	   crc(4) <= mdata(4) xor crc(22) xor crc(19) xor crc(16);
	   crc(5) <= mdata(5) xor crc(23) xor crc(21) xor crc(19) xor crc(18) xor crc(16);
	   crc(6) <= mdata(6) xor crc(22) xor crc(21) xor crc(18) xor crc(16);
	   crc(7) <= mdata(7) xor crc(23) xor crc(22) xor crc(21) xor crc(20) xor crc(18) xor crc(16);
	   crc(8) <= crc(23) xor crc(22) xor crc(21) xor crc(19) xor crc(17) xor crc(0);
	   crc(9) <= crc(23) xor crc(22) xor crc(20) xor crc(18) xor crc(1);
	   crc(10) <= crc(23) xor crc(20) xor crc(18) xor crc(17) xor crc(16) xor crc(2);
	   crc(11) <= crc(20) xor crc(16) xor crc(3);
	   crc(12) <= crc(21) xor crc(17) xor crc(4);
	   crc(13) <= crc(22) xor crc(18) xor crc(5);
	   crc(14) <= crc(23) xor crc(21) xor crc(20) xor crc(18) xor crc(17) xor crc(16) xor crc(6);
	   crc(15) <= crc(22) xor crc(21) xor crc(19) xor crc(18) xor crc(17) xor crc(7);
	   crc(16) <= crc(23) xor crc(22) xor crc(20) xor crc(19) xor crc(18) xor crc(8);
	   crc(17) <= crc(23) xor crc(18) xor crc(17) xor crc(16) xor crc(9);
	   crc(18) <= crc(21) xor crc(20) xor crc(16) xor crc(10);
	   crc(19) <= crc(22) xor crc(21) xor crc(17) xor crc(11);
	   crc(20) <= crc(23) xor crc(22) xor crc(18) xor crc(12);
	   crc(21) <= crc(23) xor crc(19) xor crc(13);
	   crc(22) <= crc(20) xor crc(14);
	   crc(23) <= crc(20) xor crc(19) xor crc(18) xor crc(17) xor crc(16) xor crc(15);
	 end if;
  end if;
end process;
d_en_conf <= '1' when maddr_cntr = "111111111111111111" or en_conf = '1' else '0';
i_en_conf : fd port map(d => d_en_conf, c => cclk, q => en_conf);
d_en_loc_bus <= '1' when (maddr_cntr(17 downto 15) = "111" and en_conf = '1') or 
                cfg_done = '1' or en_loc_bus = '1' else '0';
i_en_loc_bus : fd port map(d => d_en_loc_bus, c => cclk, q => en_loc_bus);
d_powerup <= powerup and not maddr_cntr(17);
i_powerup : fd port map(d => d_powerup, c => cclk, q => powerup);
end Behavioral;
